{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE ConstraintKinds #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE UndecidableInstances #-}

{- |
Module: Control.Monad.Reduce
-}
module Control.Monad.Reduce (
  -- * MonadReduce
  MonadReduce (..),
  (<|),
  (|>),

  -- * Combinators
  collect,

  -- * MonadReducePlus
  MonadReducePlus,
  given,
  exceptIf,
  collectNonEmpty,
  collectNonEmpty',

  -- * Helpers
  onBoth,
  liftMaybe,
  liftMaybeT,
) where

import Control.Applicative
import Control.Monad
import Control.Monad.Reader
import Control.Monad.State
import Control.Monad.Trans.Maybe
import qualified Data.List.NonEmpty as NE
import Data.Maybe
import GHC.Exception (CallStack, prettyCallStack)
import GHC.Stack (callStack, withFrozenCallStack)

-- {- | A reducer should extract itself
-- @
--  extract . red = id
-- @
-- -}
-- lawReduceId :: (MonadReduce m, Eq i) => (i -> m i) -> i -> Bool
-- lawReduceId red i = extract (red i) == i

-- | The Monad Reduce class.
class (Monad m) => MonadReduce l m | m -> l where
  {-# MINIMAL (split | check) #-}

  split :: l -> m i -> m i -> m i
  split l r1 r2 =
    check l >>= \case
      True -> r1
      False -> r2
  {-# INLINE split #-}

  -- | Check if the predicate is true.
  check :: l -> m Bool
  check l = split l (pure True) (pure False)
  {-# INLINE check #-}

-- | Infix split.
(<|) :: (MonadReduce l m, IsCallStack l) => m i -> m i -> m i
(<|) = split (fromCallStack callStack)
{-# INLINE (<|) #-}

infixr 3 <|

-- | Infix split, to the right.
(|>) :: (MonadReduce l m, IsCallStack l) => m i -> m i -> m i
r1 |> r2 = withFrozenCallStack (r2 <| r1)
{-# INLINE (|>) #-}

infixl 3 |>

class IsCallStack l where
  fromCallStack :: CallStack -> l

instance IsCallStack () where
  fromCallStack = const ()

instance IsCallStack CallStack where
  fromCallStack = id

instance IsCallStack String where
  fromCallStack = prettyCallStack

type MonadReducePlus l m = (MonadReduce l m, MonadPlus m)

instance (MonadReduce l m) => MonadReduce l (MaybeT m) where
  split l (MaybeT lhs) (MaybeT rhs) = MaybeT (split l lhs rhs)
  {-# INLINE split #-}

-- | Continues if the fact is true.
given :: (MonadReducePlus l m) => l -> m ()
given l = split l mzero (pure ())
{-# INLINE given #-}

-- | Continues if the fact is false.
exceptIf :: (MonadReducePlus l m) => l -> m ()
exceptIf l = split l mzero (pure ())
{-# INLINE exceptIf #-}

-- | Given a list of item try to remove each of them from the list.
collect :: (MonadReduce l m) => (a -> MaybeT m b) -> [a] -> m [b]
collect fn = fmap catMaybes . traverse (runMaybeT . fn)
{-# INLINE collect #-}

-- | Given a list of item try to remove each of them, but keep atleast one.
collectNonEmpty' :: (MonadReducePlus l m) => (a -> m b) -> [a] -> m [b]
collectNonEmpty' fn as =
  NE.toList <$> collectNonEmpty fn as
{-# INLINE collectNonEmpty' #-}

-- | Given a list of item try to remove each of them, but keep atleast one.
collectNonEmpty :: (MonadReducePlus l m) => (a -> m b) -> [a] -> m (NE.NonEmpty b)
collectNonEmpty fn as = do
  as' <- fmap catMaybes . traverse (optional . fn) $ as
  maybe mzero pure $ NE.nonEmpty as'
{-# INLINE collectNonEmpty #-}

-- | Helper that lifts a maybe into MonadPlus (or MaybeT)
liftMaybe :: (Alternative m) => Maybe a -> m a
liftMaybe = maybe empty pure
{-# INLINE liftMaybe #-}

liftMaybeT :: (MonadPlus m) => MaybeT m e -> m e
liftMaybeT m = runMaybeT m >>= liftMaybe
{-# INLINE liftMaybeT #-}

-- | Returns either of the maybes or combines them if both have values.
onBoth :: (MonadPlus m) => m a -> m a -> (a -> a -> m a) -> m a
onBoth mlhs mrhs fn =
  optional mlhs >>= \case
    Nothing -> mrhs
    Just lhs ->
      optional mrhs >>= \case
        Nothing -> pure lhs
        Just rhs -> fn lhs rhs
{-# INLINE onBoth #-}

instance (MonadReduce l m) => MonadReduce l (StateT s m) where
  check l = StateT (\s -> (,s) <$> check l)
  {-# INLINE check #-}

instance (MonadReduce l m) => MonadReduce l (ReaderT r m) where
  check l = ReaderT (\_ -> check l)
  {-# INLINE check #-}
