{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE DeriveDataTypeable #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Data.Expr where

import Control.Applicative
import Control.Monad.Reader
import Control.Monad.Reduce
import Control.Monad.State
import Data.Data
import qualified Data.Map.Strict as Map
import Test.Hspec

data Expr
  = Var !String
  | Cnt !Int
  | Opr !Expr !Expr
  | Let !String !Expr !Expr
  deriving (Show, Eq, Data)

isOpr :: Expr -> Bool
isOpr = \case
  Opr _ _ -> True
  _a -> False

spec :: Spec
spec = do
  describe "contains" do
    it "handles a small case" do
      contains isOpr (Opr (Var "x") (Var "y")) `shouldBe` True

contains :: (Data a) => (Expr -> Bool) -> a -> Bool
contains fn e =
  gmapQl (||) False (contains fn) e || case cast e of
    Just (e' :: Expr) -> fn e'
    Nothing -> False

rExpr
  :: (MonadReduce String m, MonadReader (Map.Map String (Either String Expr)) m)
  => Expr
  -> m Expr
rExpr e = case e of
  Cnt i -> do
    pure $ Cnt i
  Var k -> do
    v <- asks (Map.lookup k)
    case v of
      Just (Left k') -> pure $ Var k'
      Just (Right x) -> rExpr x
      Nothing -> pure (Cnt 0)
  Opr e1 e2 -> do
    split "choose left" (rExpr e1) . split "choose right" (rExpr e2) $ do
      e1' <- rExpr e1
      e2' <- rExpr e2
      case (e1', e2') of
        (Cnt n1, Cnt n2) -> do
          let opr = Opr e1' e2'
          split
            ("compute " <> prettyExprS 0 opr "")
            (pure (Cnt (n1 + n2)))
            (pure (Opr e1' e2'))
        _ow -> pure $ Opr e1' e2'
  Let k e1 e2 -> do
    e1' <- rExpr e1
    split
      ("inline " <> show k)
      (local (Map.insert k (Right e1')) $ rExpr e2)
      (Let k e1' <$> local (Map.insert k (Left k)) (rExpr e2))

modifyIn :: (Alternative m, MonadState s m) => (s -> s) -> m b -> m b
modifyIn fn mx = do
  s <- get
  put (fn s)
  x <- optional mx
  put s
  liftMaybe x

prettyExprS :: Int -> Expr -> String -> String
prettyExprS d = \case
  Var x -> showString x
  Opr l r ->
    showParen (d > addPrec) $
      prettyExprS (addPrec + 1) l
        . showString " + "
        . prettyExprS (addPrec + 1) r
  Cnt i -> shows i
  Let x e1 e2 ->
    showParen (d > letPrec) $
      showString x
        . showString " := "
        . prettyExprS (letPrec + 1) e1
        . showString "; "
        . prettyExprS letPrec e2
 where
  addPrec = 2
  letPrec = 1
